package org.rayflood.mikuvat.io;

import static org.rayflood.mikuvat.io.VSQDecoder.decode;
import static org.rayflood.mikuvat.io.VSQDecoder.encode;
import static org.rayflood.mikuvat.io.VSQDecoder.setSectionBracket;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.sound.midi.MidiEvent;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.Sequence;
import javax.sound.midi.Track;

public class VSQFile extends ImmutableObject{
	private static final long serialVersionUID = -6314152290164983853L;
	public static final String COMMON = "Common";
	public static final String MASTER = "Master";
	public static final String MIXER = "Mixer";
	public static final String EVENTLIST = "EventList";

	private File vsqFile;
	private Integer fileType;
	private Float divisionType;
	private Integer resolution;
	private MidiTrack conductor;
	private MasterSection master;
	private MixerSection mixer;
	private ArrayList<VSQTrack> tracks;

	public VSQFile(String vsq) throws Exception{
		this();
		readVSQFile(vsq);
	}

	public VSQFile(File vsq) throws Exception{
		this();
		readVSQFile(vsq);
	}

	public VSQFile(Sequence sq) throws Exception{
		this();
		setVSQSequence(sq);
	}

	public void readVSQFile(String vsq) throws Exception{
		readVSQFile(new File(vsq));
	}

	public void readVSQFile(File vsq) throws Exception{
		setVsqFile(vsq);
		setVSQSequence(MidiSystem.getSequence(getVsqFile()));
	}

	public void writeVSQFile() throws Exception{
		writeVSQFile(getVsqFile());
	}

	public void writeVSQFile(String vsq) throws Exception{
		writeVSQFile(new File(vsq));
	}

	public void writeVSQFile(File vsq) throws Exception{
		MidiSystem.write(getVSQSequence(), getFileType(), vsq);
	}

	public void setVSQSequence(Sequence sq) throws Exception{
		setFileType(MidiSystem.getMidiFileTypes(sq)[0]);
		setDivisionType(sq.getDivisionType());
		setResolution(sq.getResolution());
		Track t[] = sq.getTracks();
		setConductor(t[0]);

		Map<String, List<String>> sections = decode(t[1]);
		setMaster(new MasterSection(sections.get(MASTER)));
		setMixer(new MixerSection(sections.get(MIXER)));
		addTrack(new VSQTrack(t[1].get(0), getMixerSetting(0), sections));
		for(int i = 2; i < t.length; i++){
			addTrack(new VSQTrack(t[i].get(0), getMixerSetting(i - 1), decode(t[i])));
		}
	}

	public Sequence getVSQSequence() throws Exception{
		Sequence sq = new Sequence(getDivisionType(), getResolution());
		Track mt = sq.createTrack();
		getConductor(mt);

		List<List<String>> tsections = getVSQText();
		for(int i = 0; i < getTrackSize(); i++){
			VSQTrack vt = getTrack(i);
			Track t = sq.createTrack();
			t.add(vt.getTrackNameME());
			List<MidiEvent> me = encode(tsections.get(i));
			for(int j = 0; j < me.size(); j++){
				t.add(me.get(j));
			}
		}

		return sq;
	}

	public List<List<String>> getVSQText() throws Exception{
		if(!isReadied()){
			throw new Exception();
		}

		List<List<String>> tcommons = new ArrayList<List<String>>();
		List<List<String>> ttimelines = new ArrayList<List<String>>();
		for(int i = 0; i < getTrackSize(); i++){
			VSQTrack vt = getTrack(i);
			tcommons.add(vt.getCommon().getSection());
			ttimelines.add(vt.getTimeLine().getSectionFast());
			setMixerSetting(i, vt.getMixer());
		}

		List<List<String>> vsqtext = new ArrayList<List<String>>();
		List<String> sections = new ArrayList<String>();
		sections.add(setSectionBracket(COMMON));
		sections.addAll(tcommons.get(0));
		sections.add(setSectionBracket(MASTER));
		sections.addAll(getMaster().getSection());
		sections.add(setSectionBracket(MIXER));
		sections.addAll(getMixer().getSection());
		sections.add(setSectionBracket(EVENTLIST));
		sections.addAll(ttimelines.get(0));
		vsqtext.add(sections);

		for(int i = 1; i < getTrackSize(); i++){
			sections = new ArrayList<String>();
			sections.add(setSectionBracket(COMMON));
			sections.addAll(tcommons.get(i));
			sections.add(setSectionBracket(EVENTLIST));
			sections.addAll(ttimelines.get(i));
			vsqtext.add(sections);
		}

		return vsqtext;
	}

	public Integer[] getMixerIndexes(){
		return mixer.getIndexes();
	}

	public MixerSetting getMixerSetting(int index){
		return mixer.getMixerSetting(index);
	}

	public void setMixerSetting(int index, MixerSetting mixerSetting){
		if(mixerSetting == null || mixerSetting.isDeleted()){
			deleteMixerSetting(index);
			return;
		}
		mixer.setMixerSetting(index, mixerSetting);
	}

	public MixerSetting deleteMixerSetting(int index){
		return mixer.deleteMixerSetting(index);
	}

	public boolean hasMixerSetting(int index){
		return mixer.hasMixerSetting(index);
	}

	public int getTrackSize(){
		return tracks.size();
	}

	public void addTrack(VSQTrack vt){
		if(vt == null || vt.isDeleted()){
			return;
		}
		tracks.add(vt);
	}

	public void addTrack(int index, VSQTrack vt){
		if(vt == null || vt.isDeleted()){
			return;
		}
		if(index < getTrackSize()){
			tracks.add(index, vt);
		}
		else{
			addTrack(vt);
		}
	}

	public void setTrack(int index, VSQTrack vt){
		if(vt == null || vt.isDeleted()){
			removeTrack(index);
			return;
		}
		if(index < getTrackSize()){
			tracks.set(index, vt);
		}
		else{
			addTrack(vt);
		}
	}

	public VSQTrack getTrack(int index){
		if(index < getTrackSize()){
			return tracks.get(index);
		}
		else{
			return new VSQTrack();
		}
	}

	public VSQTrack removeTrack(int index){
		if(index < getTrackSize()){
			return tracks.remove(index);
		}
		else{
			return null;
		}
	}

	public String toString(){
		List<String> sl = new ArrayList<String>();
		if(hasVsqFile()){
			sl.add("VsqFile=" + getVsqFileS());
		}
		if(hasFileType()){
			sl.add("FileType=" + getFileType());
		}
		if(hasDivisionType()){
			sl.add("DivisionType=" + getDivisionType());
		}
		if(hasResolution()){
			sl.add("Resolution=" + getResolution());
		}
		if(hasConductor()){
			sl.add("Conductor={" + getConductor().toString() + "}");
		}
		if(hasMaster()){
			sl.add("Master={" + getMaster().toString() + "}");
		}
		if(hasMixer()){
			sl.add("Mixer={" + getMixer().toString() + "}");
		}
		if(hasTracks()){
			StringBuffer sb = new StringBuffer();
			sb.append(getTrack(0).toString());
			for(int i = 1; i < getTrackSize(); i++){
				sb.append("," + getTrack(i).toString());
			}
			sl.add("Tracks=[" + sb.toString() + "]");
		}

		StringBuffer sb = new StringBuffer();
		sb.append("VSQFile={");
		if(!sl.isEmpty()){
			sb.append(sl.get(0));
			for(int i = 1; i < sl.size(); i++){
				sb.append("," + sl.get(i));
			}
		}
		sb.append("}");
		return sb.toString();
	}

	public boolean isReadied(){
		boolean ready = true;
		ready &= hasFileType();
		ready &= hasDivisionType();
		ready &= hasResolution();
		ready &= hasConductor() && getConductor().isReadied();
		ready &= hasMaster() && getMaster().isReadied();
		ready &= hasMixer() && getMixer().isReadied();
		for(int i = 0; i < getTrackSize(); i++){
			ready &= getTrack(i).isReadied();
		}
		return ready;
	}

	public int compareTo(ImmutableObject obj){
		VSQFile vsqFile = (VSQFile)obj;
		return compare(getVsqFile().getPath(), vsqFile.getVsqFile().getPath());
	}

	public VSQFile over(Object obj, boolean fill, boolean has){
		VSQFile vsqFile = ((VSQFile)obj).clone();
		if(vsqFile.hasFileType() && (fill || has == hasFileType())){
			setFileType(vsqFile.getFileType());
		}
		if(vsqFile.hasDivisionType() && (fill || has == hasDivisionType())){
			setDivisionType(vsqFile.getDivisionType());
		}
		if(vsqFile.hasResolution() && (fill || has == hasResolution())){
			setResolution(vsqFile.getResolution());
		}
		if(vsqFile.hasConductor()){
			MidiTrack conductor = hasConductor() ? getConductor() : new MidiTrack();
			setConductor(conductor.over(vsqFile.getConductor(), fill, has));
		}
		if(vsqFile.hasMaster()){
			MasterSection master = hasMaster() ? getMaster() : new MasterSection();
			setMaster(master.over(vsqFile.getMaster(), fill, has));
		}
		if(vsqFile.hasMixer()){
			MixerSection mixer = hasMixer() ? getMixer() : new MixerSection();
			setMixer(mixer.over(vsqFile.getMixer(), fill, has));
		}
		for(int i = 0; i < vsqFile.getTracks().size(); i++){
			VSQTrack vt = getTrack(i);
			setTrack(i, vt.over(vsqFile.getTrack(i), fill, has));
		}
		if(vsqFile.hasVsqFile() && (fill || has == hasVsqFile())){
			setVsqFile(vsqFile.getVsqFile());
		}
		return this;
	}

	public VSQFile setDefault(){
		setFileType(1);
		setDivisionType(0.0f);
		setResolution(480);
		setConductor(new MidiTrack());
		setMaster(MasterSection.getDefault());
		setMixer(MixerSection.getDefault());
		deleteTracks();
		addTrack(VSQTrack.getDefault());
		deleteVsqFile();
		return this;
	}

	public static VSQFile getDefault(){
		VSQFile vsqFile = new VSQFile();
		vsqFile.setDefault();
		return vsqFile;
	}

	public VSQFile delete(){
		deleteFileType();
		deleteDivisionType();
		deleteResolution();
		deleteConductor();
		deleteMaster();
		deleteMixer();
		deleteTracks();
		deleteVsqFile();
		return this;
	}

	public boolean isDeleted(){
		boolean has = false;
		has |= hasFileType();
		has |= hasDivisionType();
		has |= hasResolution();
		has |= hasConductor();
		has |= hasMaster();
		has |= hasMixer();
		has |= hasTracks();
		has |= hasVsqFile();
		return !has;
	}

	public VSQFile(){
		delete();
	}

	public File getVsqFile(){
		return vsqFile;
	}

	public String getVsqFileS(){
		return getVsqFile().getPath();
	}

	public void setVsqFile(File vsqFile){
		this.vsqFile = vsqFile;
	}

	public void setVsqFile(String vsqFile){
		setVsqFile(new File(vsqFile));
	}

	public void deleteVsqFile(){
		vsqFile = null;
	}

	public boolean hasVsqFile(){
		return vsqFile != null;
	}

	public Integer getFileType(){
		return fileType;
	}

	public void setFileType(Integer fileType){
		this.fileType = fileType;
	}

	public void deleteFileType(){
		fileType = null;
	}

	public boolean hasFileType(){
		return fileType != null;
	}

	public Float getDivisionType(){
		return divisionType;
	}

	public void setDivisionType(Float divisionType){
		this.divisionType = divisionType;
	}

	public void deleteDivisionType(){
		divisionType = null;
	}

	public boolean hasDivisionType(){
		return divisionType != null;
	}

	public Integer getResolution(){
		return resolution;
	}

	public void setResolution(Integer resolution){
		this.resolution = resolution;
	}

	public void deleteResolution(){
		resolution = null;
	}

	public boolean hasResolution(){
		return resolution != null;
	}

	public MidiTrack getConductor(){
		return conductor;
	}

	public void getConductor(Track t) throws Exception{
		conductor.getTrack(t);
	}

	public void setConductor(MidiTrack conductor){
		if(conductor == null || conductor.isDeleted()){
			deleteConductor();
			return;
		}
		this.conductor = conductor;
	}

	public void setConductor(Track t){
		this.conductor = new MidiTrack(t);
	}

	public void deleteConductor(){
		conductor = null;
	}

	public boolean hasConductor(){
		return conductor != null;
	}

	public MasterSection getMaster(){
		return master;
	}

	public void setMaster(MasterSection master){
		if(master == null || master.isDeleted()){
			deleteMaster();
			return;
		}
		this.master = master;
	}

	public void deleteMaster(){
		master = null;
	}

	public boolean hasMaster(){
		return master != null;
	}

	public MixerSection getMixer(){
		return mixer;
	}

	public void setMixer(MixerSection mixer){
		if(mixer == null || mixer.isDeleted()){
			deleteMixer();
			return;
		}
		this.mixer = mixer;
	}

	public void deleteMixer(){
		mixer = null;
	}

	public boolean hasMixer(){
		return mixer != null;
	}

	public ArrayList<VSQTrack> getTracks(){
		return tracks;
	}

	public void setTracks(ArrayList<VSQTrack> tracks){
		if(tracks == null){
			deleteTracks();
			return;
		}
		this.tracks = tracks;
	}

	public void deleteTracks(){
		setTracks(new ArrayList<VSQTrack>());
	}

	public boolean hasTracks(){
		return !tracks.isEmpty();
	}

	public VSQFile clone(){
		return (VSQFile)super.clone();
	}

	public static VSQFile load(String file) throws Exception{
		return (VSQFile)ImmutableObject.load(file);
	}

	public static VSQFile load(File file) throws Exception{
		return (VSQFile)ImmutableObject.load(file);
	}

	public void store(String file) throws Exception{
		super.store(file);
	}

	public void store(File file) throws Exception{
		super.store(file);
	}

	public static VSQFile unzip(String file) throws Exception{
		return (VSQFile)ImmutableObject.unzip(file);
	}

	public static VSQFile unzip(File file) throws Exception{
		return (VSQFile)ImmutableObject.unzip(file);
	}

	public void zip(String file) throws Exception{
		super.zip(file);
	}

	public void zip(File file) throws Exception{
		super.zip(file);
	}
}
